#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <inttypes.h>
#include "helper.h"

uint8_t nmea_checksum(char * input){
    uint8_t idx = 0, checksum = 0, result = 0;
    //Search beginning
    while(input[idx] != '$') idx++;
    idx++;
    //Calculate checksum
    while(input[idx] != '*'){
        checksum ^= input[idx];
        idx++;
    }
    //Hexadezimal ASCII to Char
    if(input[idx+1] <= '9') result = input[idx+1] - '0';
    else result = input[idx+1] - 'A' + 10;
    result <<= 4;
    if(input[idx+2] <= '9') result |= input[idx+2] - '0';
    else result |= input[idx+2] - 'A' + 10;
    return (checksum == result);
}

nmea_GPS_dataset_p nmea_create_dataset(void){
    nmea_GPS_dataset_p result = (nmea_GPS_dataset_p)malloc(sizeof(nmea_GPS_dataset));
    result->gga = (nmea_GGA_p)malloc(sizeof(nmea_GGA));
    result->gsa = (nmea_GSA_p)malloc(sizeof(nmea_GSA));
    result->rmc = (nmea_RMC_p)malloc(sizeof(nmea_RMC));
    result->vtg = (nmea_VTG_p)malloc(sizeof(nmea_VTG));
    return result;
};

void nmea_destroy_dataset(nmea_GPS_dataset_p dataset){
    free(dataset->gga);
    free(dataset->gsa);
    free(dataset->rmc);
    free(dataset->vtg);
    free(dataset);
};

void nmea_debug_output_gga(nmea_GGA_p dataset){
    printf("Im GPGGA Datensatz von %2"PRIuFAST8":%2"PRIuFAST8":%2"PRIuFAST8" ist die Position %.4f,%c,%.4f,%c enthalten. %"PRIuFAST8" Satelliten wurden für Fix:%"PRIuFAST8" empfangen. Genauigkeit: %.1f. Höhe in %s: %.1f. WGS84-Geoid-Höhe in %s: %.1f\n",dataset->time.hour,dataset->time.minute,dataset->time.second, dataset->latitude, dataset->lat_direction, dataset->longitude, dataset->lon_direction, dataset->sattelites, dataset->fix, dataset->hdop, (dataset->altitude_unit == 'M')?"Metern":"...",dataset->altitude, (dataset->wgs84_altitude_unit == 'M')?"Metern":"...",dataset->wgs84_altitude);
};

void nmea_debug_output_gsa(nmea_GSA_p dataset) {
    printf("Im GPGSA Datensatz wurde ein %s-Fix übermittelt. Die Genauigkeiten betragen -PDOP:%.1f, -HDOP:%.1f, -VDOP:%.1f\n",(dataset->fix == 3)?"3D":((dataset->fix == 2)?"2D":"kein"),dataset->pdop,dataset->hdop,dataset->vdop);
};

void nmea_debug_output_rmc(nmea_RMC_p dataset) {
    printf("Im GPRMC Datensatz von %2"PRIuFAST8":%d:%d am %2d.%2d.%4d ist die Position %.4f,%c,%.4f,%c enthalten. Die Geschwindigkeit beträgt %.2f Knoten, der Kurs %.2f.\n",dataset->time.hour,dataset->time.minute, dataset->time.second, dataset->date.day, dataset->date.month, (dataset->date.year + 2000), dataset->latitude, dataset->lat_direction, dataset->longitude, dataset->lon_direction, dataset->speed, dataset->course);
};

void nmea_debug_output_vtg(nmea_VTG_p dataset) {
    printf("Im GPVTG Datensatz wurde der Kurs %.2f, der magn. Kurs %.2f bei einer Geschwindigkeit von %.2f Knoten (%.1f Stundenkilometer) übermittelt\n", dataset->course, dataset->mag_course, dataset->speed, dataset->speed_metric);
};

void nmea_gga_decode(char *input, nmea_GGA_p gga) {
    sscanf(input, "$GPGGA,%2"SCNuFAST8"%2"SCNuFAST8"%2"SCNuFAST8".%3"SCNuFAST16",%lf,%c,%lf,%c,%"SCNuFAST8",%"SCNuFAST8",%f,%f,%c,%f,%c,%f,%"SCNuFAST16"*",&(gga->time.hour),&(gga->time.minute),&(gga->time.second),&(gga->time.millisecond),&(gga->latitude),&(gga->lat_direction),&(gga->longitude),&(gga->lon_direction),&(gga->fix),&(gga->sattelites),&(gga->hdop),&(gga->altitude),&(gga->altitude_unit), &(gga->wgs84_altitude),&(gga->wgs84_altitude_unit),&(gga->dgps_age),&(gga->dgps_id));
};

void nmea_gsa_decode(char *input, nmea_GSA_p gsa) {
    sscanf(input, "$GPGSA,A,%"SCNuFAST8",%f,%f,%f*",&(gsa->fix),&(gsa->pdop),&(gsa->hdop),&(gsa->vdop));
};

void nmea_rmc_decode(char *input, nmea_RMC_p rmc) {
    sscanf(input, "$GPRMC,%2"SCNuFAST8"%2"SCNuFAST8"%2"SCNuFAST8".%3"SCNuFAST16",%c,%lf,%c,%lf,%c,%f,%f,%2"SCNuFAST8"%2"SCNuFAST8"%2"SCNuFAST8",%f,%c*",&(rmc->time.hour),&(rmc->time.minute),&(rmc->time.second),&(rmc->time.millisecond),&(rmc->state),&(rmc->latitude),&(rmc->lat_direction),&(rmc->longitude),&(rmc->lon_direction),&(rmc->speed),&(rmc->course),&(rmc->date.day),&(rmc->date.month),&(rmc->date.year),&(rmc->variation),&(rmc->var_direction));
};

void nmea_vtg_decode(char *input, nmea_VTG_p vtg) {
    sscanf(input, "$GPVTG,%f,T,,M,%f,N,%f,K*", &(vtg->course),&(vtg->speed),&(vtg->speed_metric));
};

void nmea_parse(char *input, nmea_GPS_dataset_p dataset){
    if(!nmea_checksum(input)){
        printf("Bad checksum!");
        return;
    }
    if(!strncmp(input, "$GPGGA", 6)) {
        nmea_gga_decode(input,dataset->gga);
        nmea_debug_output_gga(dataset->gga);
    }
    else if(!strncmp(input, "$GPGSA", 6)){
        nmea_gsa_decode(input, dataset->gsa);
        nmea_debug_output_gsa(dataset->gsa);
    }
    else if(!strncmp(input, "$GPRMC", 6)){
        nmea_rmc_decode(input, dataset->rmc);
        nmea_debug_output_rmc(dataset->rmc);
    }

    else if(!strncmp(input, "$GPVTG", 6)){
        nmea_vtg_decode(input, dataset->vtg);
        nmea_debug_output_vtg(dataset->vtg);
    }
    else {
        printf("Unsupported Foramt\n");
    }
    
};
